/*
 * RealTimer.c -- Monitor a file that contains DAY/TIME/VALUE and
 *                output the specified VALUE to the parallel port
 *                at the specified DAY/TIME.
 */

#include <windows.h>

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <string.h>
#include <time.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <error.h>
#include <malloc.h>
#include <conio.h>

#include "RealTimr.h"


#define	EVENT_FILE_EXT	"TXT"
#define CTIME_LEN		26		// ctime() string is always 26 characters including NULL
#define	CTIME_TIME		11		// ctime() time starts at offset 11


// 0 must be the first day and BAD_DAY must be the last day + 1
typedef enum { MON=0,TUE,WED,THU,FRI,SAT,SUN,BAD_DAY } days;

char *dayTab[] = {
	"Mon",
	"Tue",
	"Wed",
	"Thu",
	"Fri",
	"Sat",
	"Sun",
	"BAD DAY"
};


// local functions
void usage(void);
int readEventFile(void);
int normal(int learnFlag);
int getNow(void);
int showEvent(rtEvent_t *e,char *final);
int showList(void);
int deleteList(rtEvent_t *e);


char drive[_MAX_DRIVE];
char dir[_MAX_DIR];
char fname[_MAX_FNAME];
char ext[_MAX_EXT];

char progName[_MAX_FNAME];
int checkFlag;				// When non-zero, parse event_file and report errors
char eventFile[_MAX_PATH];
FILE *ifp;
int lineno;
int entryCnt;
FILE *ofp;
int learnFlag;

#define	DEFAULT_PORT	0x378
int portFlag;

rtEvent_t *pHead, now;	// Event head/tail and 'now' (or current) time
int mallocCnt;

int verboseFlag;		// Do more printf()s than usual


main(int argc, char *argv[])
{
	int i;
	int retVal = 0;

	_splitpath(argv[0],drive,dir,progName,ext);

	for(i=1;i<argc;++i) {
		if(*argv[i] == '-') {
			if(!strcmpi(argv[i],"-check")) {
				checkFlag = 1;
			}
			else if(!strcmpi(argv[i],"-learn")) {
				learnFlag = 1;
			}
			else if(!strcmpi(argv[i],"-verbose")) {
				verboseFlag = 1;
			}
			else if(!strnicmp(argv[i],"-port:",6)) {
				portFlag = (int)strtol(&argv[i][6],NULL,0);
			}
			else {
				retVal = -1;
				printf("ERROR:  Invalid options:  '%s'\n",argv[i]);
			}
		}
		else {
			if(*eventFile) {
				retVal = -1;
				printf("ERROR:  2nd event file found:  '%s'\n",argv[i]);
			}
			else
				strcpy(eventFile,argv[i]);
		}
	}

	if(portFlag == 0)
		portFlag = DEFAULT_PORT;

	if(verboseFlag)
		printf("Port: 0x%X\n",portFlag);

	if(eventFile[0] == '\0') {
		retVal = -1;
		printf("Error:  No event file specified on the command line!\n");
		goto Exit;
	}

	if((ifp=fopen(eventFile,"r")) == NULL) {
//		retVal = -1;
		printf("Warning:  Cannot open event file:  '%s'\n",eventFile);
		perror("");
//		goto Exit;
	}
	if(ifp != NULL) {
		fclose(ifp);
		ifp = NULL;
	}

	if((checkFlag == 0) && (learnFlag != 0)) {
		if((ofp=fopen(eventFile,"a+")) == NULL) {
			retVal = -1;
			printf("ERROR:  Cannot open event file for appending:  '%s'\n",eventFile);
			perror("");
			goto Exit;
		}
		fclose(ofp);
		ofp = NULL;
	}

	if(checkFlag) {
		if(verboseFlag) {
			getNow();
			printf("Now:  ");
			showEvent(&now,"\n");
		}
		retVal = readEventFile();
		showList();
		printf("Entries/lines:  %d/%d\n",entryCnt,lineno);
		while(pHead != NULL)
			deleteList(pHead);
		if(verboseFlag) {
			getNow();
			printf("Now:  ");
			showEvent(&now,"\n");
		}
		showList();
	}
	else if(learnFlag)
		retVal = normal(learnFlag);
	else {
		retVal = normal(learnFlag);
	}

Exit:
	if(mallocCnt)
		printf("ERROR:  allocCnt=%d\n",mallocCnt);

	exit(retVal);
}


void
usage(void)
{
	printf("usage:  %s [-options [...]] event_file\n",progName);
	printf("\n");
	printf("event_file:  File that contains DAY/TIME/VALUE lines\n");
	printf("\n");
	printf("options:\n");
	printf("          -check -- parse event_file and report errors\n");
	printf("          -learn -- press keys real time to enter on/off values\n");
	printf("          -port:val -- Parallel port data I/O address\n");
	printf("          -verbose -- Show more information than usual\n");
}


char *
getline(void)
{
	char *s;
	char *ptr;
	static char buf[256];

again:
	++lineno;
	if(fgets(buf,sizeof(buf),ifp) == NULL)
		return NULL;

	ptr = buf;
	while((*ptr == ' ') || (*ptr == '\t'))
		++ptr;

	if((*ptr == ';') || (*ptr == '\n'))
		goto again;

	for(s=ptr;*s;++s) {
		if(*s == ';') {
			*s = 0;
			break;
		}
	}

	return ptr;
}


days
getDay(char *s)
{
	if(!strnicmp(s,"mo",2))
		return MON;
	if(!strnicmp(s,"tu",2))
		return TUE;
	if(!strnicmp(s,"we",2))
		return WED;
	if(!strnicmp(s,"th",2))
		return THU;
	if(!strnicmp(s,"fr",2))
		return FRI;
	if(!strnicmp(s,"sa",2))
		return SAT;
	if(!strnicmp(s,"su",2))
		return SUN;

	return BAD_DAY;
}


int
toBcdByte(int val)
{
	int retVal;

	retVal = val / 10;
	retVal = (retVal << 4) | (val % 10);

	return retVal;
}


int
initEvent(rtEvent_t *e,char *day,char *time,int val)
{
	char *ptr;
	char buf[9];

	buf[8] = 0;	// NULL terminate string
	memcpy(buf,time,8);

	if((e->day=getDay(day)) == BAD_DAY) {
		buf[3] = 0;
		memcpy(buf,day,3);
		printf("ERROR:  Bad day value:  '%s'\n",buf);
		return -1;
	}

	e->hour = e->min = e->sec = 0;
	if((ptr=strtok(buf,":")) != NULL)
		e->hour = (int)strtol(ptr,NULL,10);
	if((ptr=strtok(NULL,":")) != NULL)
		e->min = (int)strtol(ptr,NULL,10);
	if((ptr=strtok(NULL,":")) != NULL)
		e->sec = (int)strtol(ptr,NULL,10);

	if((e->hour<0) || (e->min<0) || (e->sec<0) ||
	  (e->hour>23) || (e->min>59) || (e->sec>59)) {
		printf("ERROR:  Bad time value:  '%2d:%2d:%2d'\n",e->hour,e->min,e->sec);
		return -1;
	}

	e->dhms = e->day;
	e->dhms = (e->dhms << 8) | toBcdByte(e->hour);
	e->dhms = (e->dhms << 8) | toBcdByte(e->min);
	e->dhms = (e->dhms << 8) | toBcdByte(e->sec);

	if((e->val=val) && (val != -1))
		e->val = 0xFF;	// Turn on all 8 parallel port output bits

	return 0;
}


int
getNow(void)
{
    time_t ltime;
	char buf[CTIME_LEN];

    /* Set time zone from TZ environment variable. If TZ is not set,
     * the operating system is queried to obtain the default value 
     * for the variable.      */
    _tzset();

    /* Get UNIX-style time and display as number and string. */
    time( &ltime );
    strcpy(buf,ctime(&ltime));

	return initEvent(&now,buf,&buf[CTIME_TIME],-1);
}


int
showEvent(rtEvent_t *e,char *final)
{
#if 1
	printf("%s %2.2X:%2.2X:%2.2X",
		dayTab[e->dhms>>24],
		(e->dhms>>16)&0xFF,
		(e->dhms>> 8)&0xFF,
		(e->dhms    )&0xFF);
	if(e->val != -1)
		printf("%s",e->val ? " On" : " Off");
#else
	printf("%8.8lX",e->dhms);
#endif

	printf("%s",final?final:"");

	return 0;
}


int
appendEvent(rtEvent_t *e)
{
	int retVal = 0;

	if((ofp=fopen(eventFile,"a+")) == NULL) {
		retVal = -1;
		printf("ERROR:  Cannot open event file for appending:  '%s'\n",eventFile);
		perror("");
		goto Exit;
	}
	fprintf(ofp,"%s %2.2X:%2.2X:%2.2X %d ;; *Learned event*\n",
		dayTab[e->dhms>>24],
		(e->dhms>>16)&0xFF,
		(e->dhms>> 8)&0xFF,
		(e->dhms    )&0xFF,
		e->val);
	fclose(ofp);
	ofp = NULL;

Exit:
	return retVal;
}


int
deleteList(rtEvent_t *e)
{
	if(mallocCnt <= 0) {
		printf("ERROR:  nothing allocated!\n");
		return -1;
	}

	if(pHead == e) {
		if(e->prev->dhms < e->dhms)
			pHead = e->prev;
		else
			pHead = e->next;
	}

	e->prev->next = e->next;
	e->next->prev = e->prev;

	--mallocCnt;
	free(e);

	if(pHead == e)
		pHead = NULL;

	return 0;
}


// It's an error to have events with the same day/time value!
int
insertList(rtEvent_t *newe)
{
	rtEvent_t *prev, *next;

	if(mallocCnt <= 0) {
		printf("ERROR:  nothing allocated!\n");
		return -1;
	}

	if(pHead == NULL) {
		pHead = newe->next = newe->prev = newe;
		return 0;
	}

	prev = pHead->prev;
	while(newe->dhms > prev->next->dhms) {
		prev = prev->next;
		if(prev == pHead->prev)
			break;
	}

	if(newe->dhms == prev->dhms) {
		printf("ERROR:  two entries have the same day/time value!\n");
		return -1;
	}

	next = prev->next;

	if((pHead == next) && (newe->dhms < pHead->dhms))
		pHead = newe;

	prev->next = newe;
	next->prev = newe;
	newe->next = next;
	newe->prev = prev;

	return 0;
}


int
freeList(void)
{
	while(pHead != NULL)
		deleteList(pHead);

	return mallocCnt;
}


int
addEntry(int day,char *time,int val)
{
	int retVal = 0;
	rtEvent_t *e;

	if((e=malloc(sizeof(rtEvent_t))) == NULL) {
		perror("ERROR:  rtEvent_t malloc failed");
		return -1;
	}

	++mallocCnt;

	if((retVal=initEvent(e,dayTab[day],time,val)) != 0) {
		--mallocCnt;
		free(e);
		return retVal;
	}

	return insertList(e);
}


int
showList(void)
{
	rtEvent_t *e;

	for(e=pHead;e;) {
		showEvent(e,"\n");
		if((e=e->next) == pHead)
			break;
	}

	return 0;
}


int
readEventFile(void)
{
	int i;
	int errCnt = 0;
	char *ptr;
	char *pDay;
	char *pDay2;
	char *pTime;
	int val;
	int _day, _day2;

	if((ifp=fopen(eventFile,"r")) == NULL) {
		printf("Warning:  Cannot open event file:  '%s'\n",eventFile);
		perror("");
		return -1;
	}

	while((ptr=getline()) != NULL) {
		pDay = strtok(ptr," \t,");
		if((_day=getDay(pDay)) == BAD_DAY) {
			++errCnt;
			printf("ERROR:  Bad day:  '%s'\n",pDay);
			continue;
		}
		if((pDay2=strchr(pDay,'-')) != NULL) {
			++pDay2;
			if((_day2=getDay(pDay2)) == BAD_DAY) {
				++errCnt;
				printf("ERROR:  Bad day in range:  '%s'\n",pDay);
				continue;
			}
		}
		else
			_day2 = BAD_DAY;

		pTime = strtok(NULL," \t,");

//		strdate
		val = (int)strtol(strtok(NULL," \t,"),NULL,0);
		++entryCnt;

		if(_day2 == BAD_DAY)
			addEntry(_day,pTime,val);
		else
			for(i=_day;1;++i) {
				if(i == BAD_DAY)
					i = 0;
				addEntry(i,pTime,val);
				if(i == _day2)
					break;
			}
	}

	if(ifp != NULL) {
		fclose(ifp);
		ifp = NULL;
	}

	return errCnt;
}


rtEvent_t *
getListPtr(rtEvent_t *e)
{
	rtEvent_t *prev;

	if(pHead == NULL)
		return NULL;

	prev = pHead->prev;
	while(e->dhms > prev->next->dhms) {
		prev = prev->next;
		if(prev == pHead->prev)
			break;
	}

	return prev;
}


int
normal(int learnFlag)
{
	int ch;
	int val;
	rtEvent_t *prev, *next;
	int retVal = 0;

again:
	if((retVal=readEventFile()) != 0)
		goto Exit;

	getNow();

#if 0
	showList();
#endif

	if(((prev=getListPtr(&now)) != NULL) && (prev->val != -1)) {
		_outp((unsigned short)portFlag,val=prev->val);
		printf("EVENT:  ");
		showEvent(prev,"\n");
		printf(" next:  ");
		showEvent(prev->next,"\n");
	}
	else
		_outp((unsigned short)portFlag,val=0);

	for(;;) {
getKey:
		if(kbhit()) {
			if((ch=getch()) == 0x1B)
				goto Exit;
			if(ch == 0) {
				ch = getch();	// Function key
				while(pHead != NULL)
					deleteList(pHead);
				while(kbhit())
					getch();
				printf("\n\n*** Re-read Event file ***\n\n");
				goto again;
			}
			if(learnFlag) {
				now.val = val ? 0 : 1;
				appendEvent(&now);
				while(pHead != NULL)
					deleteList(pHead);
				while(kbhit())
					getch();
				printf("\n\n");
				goto again;
			}
			goto getKey;
		}

		getNow();
		next = getListPtr(&now);
		if(prev != next) {
			_outp((unsigned short)portFlag,val=next->val);
			printf("EVENT:  ");
			showEvent(next,"\n");
			printf(" next:  ");
			showEvent(next->next,"\n");
			prev = next;
		}
		else {
			printf("%s  ",val?"*On*: ":"*Off*:");
			showEvent(&now,"\r");
		}

		Sleep(300);
	}

Exit:
	while(pHead != NULL)
		deleteList(pHead);

	return retVal;
}
